@@ -1,20 +1,32 @@
 import logging
 import os
-import sys
 from datetime import datetime
+from io import BytesIO
 from typing import Union
 
+import azure.core.exceptions
 import pandas as pd
 from azure.storage.blob import BlobServiceClient
 from numpy import dtype
-from pandas import BooleanDtype, DataFrame, Int64Dtype, StringDtype
+from pandas import (
+    BooleanDtype,
+    CategoricalDtype,
+    DataFrame,
+    DatetimeTZDtype,
+    Float64Dtype,
+    Int8Dtype,
+    Int16Dtype,
+    Int32Dtype,
+    Int64Dtype,
+    StringDtype,
+)
 from sqlalchemy.sql.visitors import VisitableType
 from sqlalchemy.types import BigInteger, Boolean, DateTime, Integer, Numeric, String
 
 from df_to_azure.adf import ADF
-from df_to_azure.db import SqlUpsert, auth_azure, execute_stmt, test_uniqueness_columns
+from df_to_azure.db import SqlUpsert, auth_azure, execute_stmt
 from df_to_azure.exceptions import WrongDtypeError
-from df_to_azure.utils import wait_until_pipeline_is_done
+from df_to_azure.utils import test_unique_column_names, test_uniqueness_columns, wait_until_pipeline_is_done
 
 
 def df_to_azure(
@@ -30,10 +42,19 @@ def df_to_azure(
     create=False,
     dtypes=None,
     parquet=False,
+    clean_staging=True,
+    container_name="parquet",
 ):
 
     if parquet:
-        DfToParquet(df=df, tablename=tablename, folder=schema, method=method).run()
+        DfToParquet(
+            df=df,
+            tablename=tablename,
+            folder=schema,
+            method=method,
+            id_field=id_field,
+            container_name=container_name,
+        ).run()
         return None
     else:
         adf_client, run_response = DfToAzure(
@@ -48,6 +69,7 @@ def df_to_azure(
             decimal_precision=decimal_precision,
             create=create,
             dtypes=dtypes,
+            clean_staging=clean_staging,
         ).run()
 
         return adf_client, run_response
@@ -67,6 +89,7 @@ class DfToAzure(ADF):
         decimal_precision: int = 2,
         create: bool = False,
         dtypes: dict = None,
+        clean_staging: bool = True,
     ):
         super().__init__(
             df=df,
@@ -81,8 +104,14 @@ class DfToAzure(ADF):
         self.text_length = text_length
         self.decimal_precision = decimal_precision
         self.dtypes = dtypes
+        self.clean_staging = clean_staging
 
     def run(self):
+
+        if self.df.empty:
+            logging.info("Data empty, no new records to upload.")
+            return None, None
+
         if self.create:
 
             # azure components
@@ -102,6 +131,14 @@ class DfToAzure(ADF):
         run_response = self.create_pipeline(pipeline_name=self.pipeline_name)
         if self.wait_till_finished:
             wait_until_pipeline_is_done(self.adf_client, run_response)
+        if self.clean_staging & (self.method == "upsert"):
+            # If you used clean_staging=False before and the upsert gives errors on unknown columns -> remove table in
+            # staging manually
+            if not self.wait_till_finished:
+                # Only remove after pipeline is done
+                logging.info("Wait until pipeline is done before cleaning staging")
+                wait_until_pipeline_is_done(self.adf_client, run_response)
+            self.clean_staging_after_upsert()
 
         return self.adf_client, run_response
 
@@ -112,15 +149,11 @@ class DfToAzure(ADF):
 
     def upload_dataset(self):
 
-        if self.df.empty:
-            logging.info("Data empty, no new records to upload.")
-            sys.exit(1)
-
         if self.method == "create":
             self.create_schema()
             self.push_to_azure()
         if self.method == "upsert":
-            # key columns have to be unique for upsert.
+            # Key columns need only unique values for upsert
             test_uniqueness_columns(self.df, self.id_field)
             upsert = SqlUpsert(
                 table_name=self.table_name,
@@ -164,7 +197,7 @@ class DfToAzure(ADF):
             blob=f"{self.table_name}/{self.table_name}.csv",
         )
 
-        data = self.df.to_csv(index=False, sep="^", quotechar='"', line_terminator="\n")
+        data = self.df.to_csv(index=False, sep="^", quotechar='"', lineterminator="\n")
         blob_client.upload_blob(data, overwrite=True)
 
     def create_schema(self):
@@ -208,13 +241,19 @@ class DfToAzure(ADF):
             dtype("int32"): Integer(),
             dtype("int16"): Integer(),
             dtype("int8"): Integer(),
+            Int8Dtype(): Integer(),
+            Int16Dtype(): Integer(),
+            Int32Dtype(): Integer(),
             Int64Dtype(): Integer(),
+            Float64Dtype(): numeric,
             dtype("float64"): numeric,
             dtype("float32"): numeric,
             dtype("float16"): numeric,
             dtype("<M8[ns]"): DateTime(),
             dtype("bool"): Boolean(),
             BooleanDtype(): Boolean(),
+            DatetimeTZDtype(tz="utc"): DateTime(),
+            CategoricalDtype(): string,
         }
 
         col_types = {col_name: type_conversion[col_type] for col_name, col_type in self.df.dtypes.to_dict().items()}
@@ -253,6 +292,16 @@ class DfToAzure(ADF):
 
         return update_dict_bigint
 
+    def clean_staging_after_upsert(self):
+        """
+        Function to drop the table created in staging for the upsert. This function prevents issues with unmatchable
+        columns when doing upsert of different data with the same name.
+        """
+        query = f"""
+        DROP TABLE IF EXISTS staging.{self.table_name}
+        """
+        execute_stmt(query)
+
 
 class DfToParquet:
     """
@@ -261,7 +310,9 @@ class DfToParquet:
     the argument method=append to df_to_azure, the file will be created with a timestamp suffix.
     """
 
-    def __init__(self, df: pd.DataFrame, tablename: str, folder: str, method: str):
+    def __init__(
+        self, df: pd.DataFrame, tablename: str, folder: str, method: str, container_name: str, id_field: list = None
+    ):
         """
 
         Parameters
@@ -274,14 +325,27 @@ class DfToParquet:
             foldername. in df_to_azure this is the 'schema' parameter.
         method: str
             upload method (create or append supported).
+        container_name: str
+            Name of the container to write the parquet to
+        id_field: list
+            Keys to perform upsert on.
         """
 
         self.df = df
         self.tablename = tablename
-        self.upload_name = self.set_upload_name(folder, method)
+        self.method = method
+        self.id_field = id_field
+        self.upload_name = self.set_upload_name(folder)
         self.connection_string = os.environ.get("AZURE_STORAGE_CONNECTION_STRING")
+        self._checks()
+        self.container_name = container_name
+        test_unique_column_names(self.df)
 
-    def set_upload_name(self, folder: str, method: str) -> str:
+    def _checks(self):
+        if self.method == "upsert" and not self.id_field:
+            raise ValueError("With method is upsert, you need to give one or more id columns in argument id_cols")
+
+    def set_upload_name(self, folder: str) -> str:
         """
         The method parameter dictates the filename. If append is used, a timestamp will be added in order not to
         overwrite the parquet file. if method = create (default), the filename will be uploaded as is. To keep files
@@ -291,25 +355,80 @@ class DfToParquet:
         ----------
         folder: str
             the folder name, derived from the df_to_azure parameter 'schema'
-        method: str
-            upload_method.
 
         Returns
         -------
         name: str
             the folder + filename structure for uploading the dataset.
         """
-        if method == "create":
+        if self.method in ("create", "upsert"):
             name = f"{folder}/{self.tablename}.parquet"
-        elif method == "append":
+        elif self.method == "append":
             name = f"{folder}/{self.tablename}/{self.tablename}_{datetime.now().strftime('%Y%m%d%H%M%S')}.parquet"
         else:
-            raise ValueError(f"No valid method given: {method}. choose create or append.")
+            allow_list = ["create", "append", "upsert"]
+            raise ValueError(f"No valid method given: {self.method}. choose from {', '.join(allow_list)}.")
         return name
 
+    def upsert(self, df_existing: pd.DataFrame):
+        """
+        Perform insert or update on a pandas DataFrame.
+
+        The new values of self.df will be updated in df_existing,
+        and new rows in self.df will be appended to df_existing.
+
+        1. We read the existing parquet file from storage
+        2. We do the upsert with the given dataframe
+        3. We upload and overwrite the parquet file on storage with the updated dataframe
+
+        Parameters
+        ----------
+        df_existing: pd.DataFrame
+            The dataframe which is already on Azure Storage and has to be updated.
+
+        Returns
+        -------
+        result: pd.DataFrame
+            Updated dataframe to be uploaded.
+        """
+        # check if columns are equal
+
+        diff_cols = self.df.columns.symmetric_difference(df_existing.columns)
+        if diff_cols.any():
+            err_msg = (
+                f"When performing upsert, column names must be equal. " f"Difference in columns: {', '.join(diff_cols)}"
+            )
+            raise ValueError(err_msg)
+
+        # check if there are any NaNs in the new data, else we use pd.concat instead of combine_first
+        idx = self.df.set_index(self.id_field)
+        any_nan = idx.isna().any(axis=1)
+        if any_nan.any():
+            self.df = pd.concat([self.df, df_existing]).drop_duplicates(subset=self.id_field)
+            self.df = self.df.sort_values(self.id_field, ignore_index=True)
+        else:
+            # set indices with id columns
+            df_existing = df_existing.set_index(self.id_field)
+            # perform upsert
+            self.df = idx.combine_first(df_existing)
+            # set id columns back as regular colums
+            self.df = self.df.reset_index()
+
     def run(self):
+        blob_service_client = BlobServiceClient.from_connection_string(self.connection_string)
+        container_client = blob_service_client.get_container_client(container=self.container_name)
+
+        if self.method == "upsert":
+            test_uniqueness_columns(self.df, self.id_field)
+            downloaded_blob = container_client.download_blob(self.upload_name)
+            bytes_io = BytesIO(downloaded_blob.readall())
+            df_existing = pd.read_parquet(bytes_io)
+            self.upsert(df_existing=df_existing)
 
         text_stream = self.df.to_parquet()
-        blob_service_client = BlobServiceClient.from_connection_string(self.connection_string)
-        container_client = blob_service_client.get_container_client(container="parquet")
-        container_client.upload_blob(data=text_stream, name=self.upload_name, overwrite=True)
+        try:
+            container_client.upload_blob(data=text_stream, name=self.upload_name, overwrite=True)
+        except azure.core.exceptions.ResourceNotFoundError:
+            logging.info(f"Container {self.container_name} is created!")
+            container_client.create_container()
+            container_client.upload_blob(data=text_stream, name=self.upload_name, overwrite=True)
